==> clickhouse-connect-0.0.11.tar.gz: clickhouse-connect-0.0.11/clickhouse_connect/dbapi/cursor.py <==
import logging

from typing import Optional, Sequence

from clickhouse_connect.datatypes.numeric import Int32
from clickhouse_connect.datatypes.string import String
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver import Client
from clickhouse_connect.driver.parser import parse_callable


class Cursor:
    """
    See :ref:`https://peps.python.org/pep-0249/`
    """
    def __init__(self, client: Client):
        self.client = client
        self.arraysize = 1
        self.data: Optional[Sequence] = None
        self.names = []
        self.types = []
        self._rowcount = 0
        self._ix = 0

    def check_valid(self):
        if self.data is None:
            raise ProgrammingError('Cursor is not valid')

    @property
    def description(self):
        return [(n, t.name, None, None, None, None, True) for n, t in zip(self.names, self.types)]

    @property
    def rowcount(self):
        return self._rowcount

    def close(self):
        self.data = None

    def execute(self, operation:str, parameters=None):
        op_start = operation.split(' ')[0].upper()
        if op_start in ('SELECT', 'WITH', 'SHOW', 'DESCRIBE', 'EXISTS'):
            query_result = self.client.query(operation, parameters)
            self.data = query_result.result_set
            self.names = query_result.column_names
            self.types = query_result.column_types
        else:
            v = self.client.command(operation, parameters)
            self.types = [String]
            self.names = ['response']
            if not v:
                v = 'OK'
            elif isinstance(v, int):
                self.types = [Int32]
            if isinstance(v, list):
                v = '\t'.join(v)
            self.data = [[v]]
        self._rowcount = len(self.data)

    def _try_bulk_insert(self, operation:str, data):
        if not operation.upper().startswith('INSERT INTO '):
            return False
        temp = operation[11:].strip()
        table_end = min(temp.find(' '), temp.find('('))
        table = temp[:table_end].strip()
        temp = temp[table_end:].strip()
        if temp[0] == '(':
            _, op_columns, temp = parse_callable(temp)
        else:
            op_columns = None
        if 'VALUES' not in temp.upper():
            return False
        col_names = list(data[0].keys())
        if op_columns and set(op_columns) != set(col_names):
            return False  # Data sent in doesn't match the columns in the insert statement
        if self.client.column_inserts:
            data_values = [[row[k] for row in data] for k in col_names]
        else:
            data_values = [list(row.values()) for row in data]
        self.client.insert(table, data_values, col_names, column_oriented=self.client.column_inserts)
        self.data = []
        return True

    def executemany(self, operation, parameters):
        if not parameters or self._try_bulk_insert(operation, parameters):
            return
        self.data = []
        try:
            for param_row in parameters:
                query_result = self.client.query(operation, param_row)
                self.data.extend(query_result.result_set)
                if self.names or self.types:
                    if query_result.column_names != self.names:
                        logging.warning('Inconsistent column names %s : %s for operation %s in cursor executemany',
                                        self.names, query_result.column_names, operation)
                else:
                    self.names = query_result.column_names
                    self.types = query_result.column_types
        except TypeError as ex:
            raise ProgrammingError(f'Invalid parameters {parameters} passed to cursor executemany') from ex
        self._rowcount = len(self.data)

    def fetchall(self):
        self.check_valid()
        ret = self.data
        self._ix = self._rowcount
        return ret

    def fetchone(self):
        self.check_valid()
        if self._ix >= self._rowcount:
            return None
        val = self.data[self._ix]
        self._ix += 1
        return val

    def fetchmany(self, size: int = -1):
        self.check_valid()
        end = self._ix + max(size, self._rowcount - self._ix)
        ret = self.data[self._ix: end]
        self._ix = end
        return ret

    def nextset(self):
        raise NotImplementedError

    def callproc(self, *args, **kwargs):
        raise NotImplementedError

==> clickhouse-connect-0.0.14.tar.gz: clickhouse-connect-0.0.14/clickhouse_connect/dbapi/cursor.py <==
import logging

from typing import Optional, Sequence

from clickhouse_connect.datatypes.numeric import Int32
from clickhouse_connect.datatypes.string import String
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver import Client
from clickhouse_connect.driver.parser import parse_callable

logger = logging.getLogger(__name__)


class Cursor:
    """
    See :ref:`https://peps.python.org/pep-0249/`
    """
    def __init__(self, client: Client):
        self.client = client
        self.arraysize = 1
        self.data: Optional[Sequence] = None
        self.names = []
        self.types = []
        self._rowcount = 0
        self._ix = 0

    def check_valid(self):
        if self.data is None:
            raise ProgrammingError('Cursor is not valid')

    @property
    def description(self):
        return [(n, t.name, None, None, None, None, True) for n, t in zip(self.names, self.types)]

    @property
    def rowcount(self):
        return self._rowcount

    def close(self):
        self.data = None

    def execute(self, operation:str, parameters=None):
        op_start = operation.split(' ')[0].upper()
        if op_start in ('SELECT', 'WITH', 'SHOW', 'DESCRIBE', 'EXISTS'):
            query_result = self.client.query(operation, parameters)
            self.data = query_result.result_set
            self.names = query_result.column_names
            self.types = query_result.column_types
        else:
            v = self.client.command(operation, parameters)
            self.types = [String]
            self.names = ['response']
            if not v:
                v = 'OK'
            elif isinstance(v, int):
                self.types = [Int32]
            if isinstance(v, list):
                v = '\t'.join(v)
            self.data = [[v]]
        self._rowcount = len(self.data)

    def _try_bulk_insert(self, operation:str, data):
        if not operation.upper().startswith('INSERT INTO '):
            return False
        temp = operation[11:].strip()
        table_end = min(temp.find(' '), temp.find('('))
        table = temp[:table_end].strip()
        temp = temp[table_end:].strip()
        if temp[0] == '(':
            _, op_columns, temp = parse_callable(temp)
        else:
            op_columns = None
        if 'VALUES' not in temp.upper():
            return False
        col_names = list(data[0].keys())
        if op_columns and set(op_columns) != set(col_names):
            return False  # Data sent in doesn't match the columns in the insert statement
        if self.client.column_inserts:
            data_values = [[row[k] for row in data] for k in col_names]
        else:
            data_values = [list(row.values()) for row in data]
        self.client.insert(table, data_values, col_names, column_oriented=self.client.column_inserts)
        self.data = []
        return True

    def executemany(self, operation, parameters):
        if not parameters or self._try_bulk_insert(operation, parameters):
            return
        self.data = []
        try:
            for param_row in parameters:
                query_result = self.client.query(operation, param_row)
                self.data.extend(query_result.result_set)
                if self.names or self.types:
                    if query_result.column_names != self.names:
                        logger.warning('Inconsistent column names %s : %s for operation %s in cursor executemany',
                                        self.names, query_result.column_names, operation)
                else:
                    self.names = query_result.column_names
                    self.types = query_result.column_types
        except TypeError as ex:
            raise ProgrammingError(f'Invalid parameters {parameters} passed to cursor executemany') from ex
        self._rowcount = len(self.data)

    def fetchall(self):
        self.check_valid()
        ret = self.data
        self._ix = self._rowcount
        return ret

    def fetchone(self):
        self.check_valid()
        if self._ix >= self._rowcount:
            return None
        val = self.data[self._ix]
        self._ix += 1
        return val

    def fetchmany(self, size: int = -1):
        self.check_valid()
        end = self._ix + max(size, self._rowcount - self._ix)
        ret = self.data[self._ix: end]
        self._ix = end
        return ret

    def nextset(self):
        raise NotImplementedError

    def callproc(self, *args, **kwargs):
        raise NotImplementedError
